import warnings
from mqt import *

warnings.simplefilter('always', DeprecationWarning)
warnings.warn('Usage via `import jkq` is deprecated in favor of the new prefix. Please use `import mqt` instead.', DeprecationWarning)
